@@ -1,18 +1,25 @@
 #!/usr/bin/python
-# SECUREAUTH LABS. Copyright 2018 SecureAuth Corporation. All rights reserved.
+# Impacket - Collection of Python classes for working with network protocols.
 #
-# This software is provided under under a slightly modified version
+# Copyright Fortra, LLC and its affiliated companies 
+#
+# All rights reserved.
+#
+# This software is provided under a slightly modified version
 # of the Apache Software License. See the accompanying LICENSE file
 # for more information.
 #
-# Description: DCE/RPC SAMR dumper.
+# Description:
+#   DCE/RPC SAMR dumper.
 #
 # Author:
-#  Javier Kohen <jkohen@coresecurity.com>
-#  Alberto Solino (@agsolino)
+#   Javier Kohen
+#   Alberto Solino (@agsolino)
 #
 # Reference for:
-#  DCE/RPC for SAMR
+#   DCE/RPC for SAMR
+#
+
 from __future__ import division
 from __future__ import print_function
 import sys
@@ -84,7 +91,7 @@ class SAMRDump:
         # Display results.
 
         if self.__csvOutput is True:
-            print('#Name,RID,FullName,PrimaryGroupId,BadPasswordCount,LogonCount,PasswordLastSet,PasswordDoesNotExpire,AccountIsDisabled,UserComment,ScriptPath')
+            print('#Name,RID,FullName,PrimaryGroupId,BadPasswordCount,LogonCount,PasswordLastSet,PasswordDoesNotExpire,AccountIsDisabled,AdminComment,UserComment,ScriptPath')
 
         for entry in entries:
             (username, uid, user) = entry
@@ -105,13 +112,14 @@ class SAMRDump:
                 accountDisabled = 'False'
 
             if self.__csvOutput is True:
-                print('%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s' % (username, uid, user['FullName'], user['PrimaryGroupId'],
+                print('%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s' % (username, uid, user['FullName'], user['PrimaryGroupId'],
                                                       user['BadPasswordCount'], user['LogonCount'],pwdLastSet,
-                                                      dontExpire, accountDisabled, user['UserComment'].replace(',','.'),
+                                                      dontExpire, accountDisabled, user['UserComment'].replace(',','.'),user['AdminComment'].replace(',','.'),
                                                       user['ScriptPath']  ))
             else:
                 base = "%s (%d)" % (username, uid)
                 print(base + '/FullName:', user['FullName'])
+                print(base + '/AdminComment:', user['AdminComment'])
                 print(base + '/UserComment:', user['UserComment'])
                 print(base + '/PrimaryGroupId:', user['PrimaryGroupId'])
                 print(base + '/BadPasswordCount:', user['BadPasswordCount'])
@@ -229,14 +237,7 @@ if __name__ == '__main__':
     options = parser.parse_args()
 
     # Init the example's logger theme
-    logger.init(options.ts)
-
-    if options.debug is True:
-        logging.getLogger().setLevel(logging.DEBUG)
-        # Print the Library's installation path
-        logging.debug(version.getInstallationPath())
-    else:
-        logging.getLogger().setLevel(logging.INFO)
+    logger.init(options.ts, options.debug)
 
     domain, username, password, remoteName = parse_target(options.target)
 
